(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import list_fall_exst.
Require Import predicate.

Require Import labels.
Require Import formula.
Require Import words.
Require Import constraints.
Require Import pme.
Require Import css.

Section BBI_expansion.

  Variable p : (formula*formula*label*label*word label*word label*word label)%type.
  
  Let A := match p with (A,_,_,_,_,_,_) => A end.
  Let B := match p with (_,B,_,_,_,_,_) => B end.
  Let a := match p with (_,_,a,_,_,_,_) => a end.
  Let b := match p with (_,_,_,b,_,_,_) => b end.
  Let m := match p with (_,_,_,_,m,_,_) => m end.
  Let x := match p with (_,_,_,_,_,x,_) => x end.
  Let y := match p with (_,_,_,_,_,_,y) => y end.
  
  Variable css : CSS_type.
  
  Let X := match css with (X,_) => X end.
  Let C := match css with (_,C) => C end.

  Notation ec := (Emptyset cst).
  Notation es := (Emptyset stm).
  Notation eps := (nil : word label).
  
  Definition bbi_exp_T_not_cond := X (T (!A) # m).
  Definition bbi_exp_T_not_inst := (sg (F A # m), ec)::nil.
  
  Definition bbi_exp_F_not_cond := X (F (!A) # m).
  Definition bbi_exp_F_not_inst := (sg (T A # m), ec)::nil.
  
  Definition bbi_exp_T_and_cond := X (T A && B # m).
  Definition bbi_exp_T_and_inst := (sg (T A # m) cup sg (T B # m), ec)::nil.
  
  Definition bbi_exp_F_and_cond := X (F A && B # m).
  Definition bbi_exp_F_and_inst := (sg (F A # m), ec)::(sg (F B # m), ec)::nil.
  
  Definition bbi_exp_T_unit_cond := X (T I # m).
  Definition bbi_exp_T_unit_inst := (es, sg (eps~~m))::nil.
  
  Definition bbi_exp_T_mult_cond := X (T (A * B) # m) /\ a <> b /\ ~ alphabet C a /\ ~ alphabet C b.
  Definition bbi_exp_T_mult_inst := (sg (T A # l a) cup sg (T B # l b), sg (l a o l b ~~ m))::nil.
  
  Definition bbi_exp_F_mult_cond := X (F (A * B) # m) /\ pme C (x o y ~~ m).
  Definition bbi_exp_F_mult_inst := (sg (F A # x), ec)::(sg (F B # y), ec)::nil.
  
  Definition bbi_exp_T_wand_cond := X (T (A -* B) # m) /\ pme C (x o m ~~ y).
  Definition bbi_exp_T_wand_inst := (sg (F A # x), ec)::(sg (T B # y), ec)::nil.
  
  Definition bbi_exp_F_wand_cond := X (F (A -* B) # m) /\ a <> b /\ ~ alphabet C a /\ ~ alphabet C b.
  Definition bbi_exp_F_wand_inst := (sg (T A # l a) cup sg (F B # l b), sg (l a o m ~~ l b))::nil.
  
End BBI_expansion.

Record bbi_tab_rule : Type := MkBBI_rule {
  bbi_param : Type;
  bbi_cond  : bbi_param -> CSS_type -> Prop;
  bbi_inst  : bbi_param -> list CSS_type
}.

Implicit Arguments MkBBI_rule [ bbi_param ].

Definition bbi_exp_T_not  := MkBBI_rule bbi_exp_T_not_cond  bbi_exp_T_not_inst.
Definition bbi_exp_F_not  := MkBBI_rule bbi_exp_F_not_cond  bbi_exp_F_not_inst.
Definition bbi_exp_T_and  := MkBBI_rule bbi_exp_T_and_cond  bbi_exp_T_and_inst.
Definition bbi_exp_F_and  := MkBBI_rule bbi_exp_F_and_cond  bbi_exp_F_and_inst.
Definition bbi_exp_T_unit := MkBBI_rule bbi_exp_T_unit_cond bbi_exp_T_unit_inst.
Definition bbi_exp_T_mult := MkBBI_rule bbi_exp_T_mult_cond bbi_exp_T_mult_inst.
Definition bbi_exp_F_mult := MkBBI_rule bbi_exp_F_mult_cond bbi_exp_F_mult_inst.
Definition bbi_exp_T_wand := MkBBI_rule bbi_exp_T_wand_cond bbi_exp_T_wand_inst.
Definition bbi_exp_F_wand := MkBBI_rule bbi_exp_F_wand_cond bbi_exp_F_wand_inst.

Definition bbi_exp_rules := 
    bbi_exp_T_not::bbi_exp_F_not::bbi_exp_T_and::bbi_exp_F_and::
    bbi_exp_T_unit::bbi_exp_T_mult::bbi_exp_F_mult::bbi_exp_T_wand::bbi_exp_F_wand::nil.

Definition bbi_expand_1 (css1 css2 : CSS_type) := match (css1,css2) with ((X,C),(X1,C1)) => (X cup X1, C cup C1) end.
Definition bbi_expand CSS lCSS := map (bbi_expand_1 CSS) lCSS.

Notation bbi_fireable := bbi_cond.

Inductive bbi_expansion css : list CSS_type -> Prop := 
  | In_bbi_exp : forall r p,      In r bbi_exp_rules 
                               -> bbi_fireable r p css
                               -> bbi_expansion css (bbi_expand css (bbi_inst r p)).

Fixpoint bbi_tab_constraints (lcss : list CSS_type) :=
  match lcss with
    | nil         => empty
    | (X,C)::lcss => C cup (bbi_tab_constraints lcss)
  end.

Section BBI_exp_basic.

  Variable X : stm -> Prop.
  Variable C : cst -> Prop.

  Ltac solve r1 p1 := 
    intros;
    apply In_bbi_exp with (css := (X,C)) (r := r1) (p := p1); 
    [ unfold bbi_exp_rules; In_auto
    | simpl; auto ]; red; auto.
  
  Proposition bbi_exp_Tnot A m : X (T !A # m) -> bbi_expansion (X,C) ((X cup sg (F A # m), C cup empty)::nil).
  Proof. solve bbi_exp_T_not (A,A,l0,l0,m,m,m). Qed.
  
  Proposition bbi_exp_Fnot A m : X (F !A # m) -> bbi_expansion (X,C) ((X cup sg (T A # m), C cup empty)::nil).
  Proof. solve bbi_exp_F_not (A,A,l0,l0,m,m,m). Qed.
 
  Proposition bbi_exp_Tand A B m : X (T A && B # m) -> bbi_expansion (X,C) ((X cup (sg (T A # m) cup sg (T B # m)), C cup empty)::nil).
  Proof. solve bbi_exp_T_and (A,B,l0,l0,m,m,m). Qed.
  
  Proposition bbi_exp_Fand A B m : X (F A && B # m) -> bbi_expansion (X,C) ((X cup sg (F A # m), C cup empty)::
                                                                                    (X cup sg (F B # m), C cup empty)::nil).
  Proof. solve bbi_exp_F_and (A,B,l0,l0,m,m,m). Qed.

  Proposition bbi_exp_TI m : X (T I # m) -> bbi_expansion (X,C) ((X cup empty, C cup sg (eps ~~ m))::nil).
  Proof. solve bbi_exp_T_unit (I,I,l0,l0,m,m,m). Qed.

  Proposition bbi_exp_Tmult A B a b m : X (T A * B # m) -> a<> b -> ~ alphabet C a -> ~ alphabet C b 
                                             -> bbi_expansion (X,C) ((X cup (sg (T A # l a) cup sg (T B # l b)), C cup sg (l a o l b ~~ m))::nil).
  Proof. solve bbi_exp_T_mult (A,B,a,b,m,m,m). Qed.

  Proposition bbi_exp_Fmult A B x y m : X (F A * B # m) -> pme C (x o y ~~ m)
                                             -> bbi_expansion (X,C) ((X cup sg (F A # x), C cup empty)::
                                                                     (X cup sg (F B # y), C cup empty)::nil).
  Proof. solve bbi_exp_F_mult (A,B,l0,l0,m,x,y). Qed.

  Proposition bbi_exp_Twand A B x y m : X (T A -* B # m) -> pme C (x o m ~~ y)
                                             -> bbi_expansion (X,C) ((X cup sg (F A # x), C cup empty)::
                                                                     (X cup sg (T B # y), C cup empty)::nil).
  Proof. solve bbi_exp_T_wand (A,B,l0,l0,m,x,y). Qed.

  Proposition bbi_exp_Fwand A B a b m : X (F A -* B # m) -> a<> b -> ~ alphabet C a -> ~ alphabet C b 
                                             -> bbi_expansion (X,C) ((X cup (sg (T A # l a) cup sg (F B # l b)), C cup sg (l a o m ~~ l b))::nil).
  Proof. solve bbi_exp_F_wand (A,B,a,b,m,m,m). Qed.

End BBI_exp_basic.
