(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Arith.
Require Import NPeano.

(* The Soundness and Completeness of Kripke semantics cannot be proved without classical logic *)

Require Import Classical.
Require Import classical_help.

Require Import predicate.
Require Import closure.
Require Import finite.
Require Import carrier.
Require Import enum.
Require Import maps.
 
Require Import formula.
Require Import labels.
Require Import words.
Require Import constraints.
Require Import pme.
Require Import css.

Require Import kripke.

Require Import tableaux.

Require Import bbi_expansion.
Require Import bbi_exp_props.
Require Import bbi_closed_branch.
Require Import bbi_tableaux.
Require Import bbi_realizability.

Require Import bbi_strategy.
Require Import bbi_oracle.

Require Import bbi_simple.
Require Import bbi_basic.
Require Import bbi_limit.
Require Import bbi_hintikka.

Definition bbi_tableau_provable f   := exists a, bbi_has_closed_tableau (sg (F f # l a), sg (l a ~~ l a)).
Definition bbi_validity f           := forall L (meq : constraint L -> Prop) fV, bbi_frame meq fV -> mdl_validity meq fV f.
Definition bbi_counter_model {L} (meq : constraint L -> Prop) fV m f := bbi_frame meq fV /\ meq (m~~m) /\ ~ forces meq fV m f.

Proposition CSS_FAab A x y : finite_CSS (sg (F A # x), sg (x ~~ y)).
Proof. 
  split; auto.
  intros s f' m' H. 
  apply in_sg_eq in H. injection H. 
  intros Ha0 Hf' Hs.
  rewrite -> Ha0.
  apply pme_lft with (y := y), pme_inc; auto.
Qed.

Hint Resolve CSS_FAab.

Theorem bbi_tab_soundness f a b : bbi_has_closed_tableau (sg (F f # l a), sg (l a ~~ l b)) -> bbi_validity f. 
Proof.
  apply contraposition.
  intros Hf; unfold bbi_validity, mdl_validity in Hf.
  apply not_all_ex_not in Hf.
  destruct Hf as [ L Hf ]; simpl in Hf.
  classic_solve.
  destruct Hf as (meq & fV & Hframe & Hf).
  classic_solve.
  destruct Hf as ( x & Hx & Hf ).

  set ( b0 := (sg (F f # l a), sg (l a ~~ l b)) ).
  assert (CSS b0) as HCSS. unfold b0; apply CSS_FAab; auto.
  assert (realizable b0) as Real.
    red. split. trivial.
    unfold b0.
    exists L.
    exists meq.
    exists fV.
    exists (fun _ => x). 
    split; auto.
    red. split; auto.
    split; pred_solve; simpl; rewrite -> eps_neutral_r; trivial.
  apply realizable_has_no_closed_tableau with (2 := realizable_stable); auto.
  apply realizable_id.
  apply realizable_not_closed.
Qed.

Corollary bbi_tableau_soundness : bbi_tableau_provable inc bbi_validity.
Proof. intros f [ ? H ] M; apply bbi_tab_soundness with (1 := H); auto. Qed.

Corollary bbi_tableau_soundness' f L (meq : constraint L -> Prop) fV m : bbi_frame meq fV -> bbi_tableau_provable f -> meq (m~~m) -> forces meq fV m f.
Proof. intros; apply bbi_tableau_soundness; auto. Qed.

Section Hintikka_counter_model.

  Variable X : stm -> Prop.
  Variable C : cst -> Prop.
  Hypothesis HXC : Hintikka X C.
  Variable f : formula.
  Variable m : word label.
  Hypothesis Hf : X (F f # m).
  
  Corollary Hintikka_counter_model : bbi_counter_model (pme C) (bbi_herbrand X C) m f.
  Proof.
    split; auto.
    apply bbi_herbrand_prop.
    split.
    destruct HXC as [ H ].
    apply H in Hf; auto.
    apply (proj2 (Hintikka_model _ _ HXC)) in Hf.
    simpl in Hf; rewrite subst_w_id  in Hf; auto.
  Qed.

End Hintikka_counter_model.

Section Tableau_strong_completeness.

  Variable a b   : label.
  Hypothesis Hab : a <> b.
  Variable f     : formula.
  Hypothesis Hf  : bbi_has_no_closed_tableau (sg (F f # l a), sg (l a ~~ l b)).

  Let c := idx_dmap _ idx_label.
  Let na := idx_rmap _ idx_label a.
  Let nb := idx_rmap _ idx_label b.
  Let n0 := S (max na nb).
  Let X0 := sg (F f # l a).
  Let C0 := sg (eps ~~ eps) cup sg (l a ~~ l b).

  Let istm := projT1 fair_strategy_exists.
  Let Histm : fair_strategy istm.
  Proof. exact (projT2 fair_strategy_exists). Qed.

  Let P := projT1 oracle_exists.
  Let OP : oracle P.
  Proof. apply (projT2 oracle_exists). Qed.
  Let HP : CSS_consistent inc P.
  Proof. apply (projT2 oracle_exists). Qed.

  Let Pdec : (forall css, P css \/ ~ P css).
  Proof. intros; apply classic. Qed.

  Let HXC0 : finite_CSS (X0,C0).
  Proof.
    unfold X0, C0.
    split; auto.
    apply CSS_sg_l.
    apply pme_lft with (y := l b), pme_inc; pred_solve.
  Qed.
  
  Let H0 : C0 (eps ~~ eps).
  Proof. unfold C0; pred_solve. Qed.
    
  Let H1 : CSS_consistent (X0,C0).
  Proof.
    unfold CSS_consistent, X0, C0.
    apply CSS_consistent_closed with (4 := Hf); auto.
  Qed.

  Let H2 : alphabet C0 inc pd_image c (fun n => n < n0).
  Proof.
    unfold C0.
    apply inc_transitive with (1 := proj1 (alphabet_cup _ _)).
    pred_solve.
    apply inc_transitive with (1 := @alphabet_sg_eps _); auto.
    apply inc_transitive with (1 := proj1 (alphabet_sg _ _)).
    pred_solve; apply carrier_cons_2; auto; unfold n0.
    exists na. split.
    generalize (Nat.le_max_l na nb); omega.
    unfold na; apply idx_surj.
    exists nb. split.
    generalize (Nat.le_max_r na nb); omega.
    unfold nb; apply idx_surj.
  Qed.

  Let HXa : X0 (F f # l a).
  Proof. unfold X0; auto. Qed.

  Let HC0 : bbi_basic_pme (pme C0).
  Proof.
    unfold C0.
    constructor 2 with (R := pme (sg (eps~~eps))) (c := l a ~~ l b).
    cutrewrite (l a = l a o eps).
    constructor 2; auto.
    intros H.
    apply alphabet_pme in H.
    apply alphabet_sg_eps in H.
    destruct H.
    intros H.
    apply alphabet_pme in H.
    apply alphabet_sg_eps in H.
    destruct H.
    auto.
    constructor 1.
    split; auto.
    apply cl_cl_cup_simplify; auto.
  Qed.

  Definition Xcm := lX istm P n0 X0 C0.
  Definition Ccm := lC istm P n0 X0 C0.

  Let G0 : bbi_simple_pme (pme Ccm).
  Proof. exact (lC_simple istm _ Pdec HP OP n0 _ _ H0 H1 H2 HC0). Qed.

  Let G1 : CSS_leq (X0,C0) (Xcm,Ccm).
  Proof. exact (XC0_inc_lXC istm _ Pdec n0 X0 C0). Qed.

  Let G2 : Hintikka Xcm Ccm.
  Proof. exact (lXC_Hintikka _ Histm _ Pdec HP OP n0 _ _ H0 H1 H2). Qed.   

  Theorem bbi_tab_strong_completeness : bbi_counter_model (pme Ccm) (bbi_herbrand Xcm Ccm) (l a) f /\ bbi_simple_pme (pme Ccm).
  Proof.
    split; auto.
    apply Hintikka_counter_model; auto.
    apply G1; auto.
  Qed.

End Tableau_strong_completeness.

Theorem bbi_tableau_strong_completeness f : ~ bbi_tableau_provable f 
                                         -> { L : _ & { meq : constraint L -> Prop & { fV : _ & { m | 
                                              bbi_counter_model meq fV m f /\ bbi_simple_pme meq 
                                            } } } }.  
Proof.
  intros Hf.
  set (a := idx_dmap _ idx_label 0).
  set (b := idx_dmap _ idx_label 1).
  exists label.
  exists (pme (Ccm a b f)).
  exists (bbi_herbrand (Xcm a b f) (Ccm a b f)).
  exists (l a).
  apply bbi_tab_strong_completeness.
  intros H0; apply indexation_dinj in H0; discriminate H0.
  split; auto.
  contradict Hf.
  apply bbi_has_proof_aa in Hf.
  exists a; auto.
Qed.

Theorem bbi_tableau_completeness : bbi_validity inc bbi_tableau_provable. 
Proof.
  intros f H; destruct (classic (bbi_tableau_provable f)) as [ | H' ]; auto.
  destruct (bbi_tableau_strong_completeness _ H') as (? & ? & ? & ? & (H1 & H2 & H3) & _).
  contradiction (H3 (H _ _ _ H1 _ H2)).
Qed.