(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.

Require Import predicate.
Require Import maps.
Require Import natpred.
Require Import natseq.
Require Import maxf.

Section enum_P.

  (* Compute the increasing enumeration of a infinite and decidable predicate on nat *)

  Variable P : nat -> Prop.
  Variable Pdec : pnat_dec P.
  Variable Pinf : pnat_inf P.

  Let pn := pnat_next _ Pdec Pinf.  

  Fixpoint enum_P n := match n with 
    | 0   => proj1_sig (pn 0)
    | S n => proj1_sig (pn (S (enum_P n)))
  end.

  Theorem enum_P_sinc : fnat_sinc enum_P.
  Proof.
    apply fnat_inc_imp_sinc; intros x.
    generalize (proj2_sig (pn (S (enum_P x)))).
    intros [ H [ ? ? ] ]; apply H.
  Qed.

  Proposition enum_P_prop1 : forall x, P (enum_P x).
  Proof.
    induction x; simpl;
    match goal with [ |- P (proj1_sig ?k) ] => generalize (proj2_sig k); tauto end.
  Qed.

  Proposition enum_P_prop2 : forall k, k < enum_P 0 -> ~ P k.
  Proof.
    intros k Hk. simpl in Hk.
    generalize (proj2_sig (pn 0)). 
    intros [ H1 [ H2 H3 ] ].
    destruct (Pdec k) as [ H | H ]; auto.
    generalize (H3 _ (le_O_n _) H). 
    intros C; contradict C; auto with arith.
  Qed.
   
  Proposition enum_P_prop3 : forall x k, enum_P x < k -> k < enum_P (S x) -> ~ P k.
  Proof.
    intros x k Hx Hk.
    simpl in Hk.
    generalize (proj2_sig (pn (S (enum_P x)))).
    intros [ H1 [ H2 H3 ] ].
    destruct (Pdec k) as [ H | H ]; auto.
    generalize (H3 _ Hx H).
    intros C; contradict C; auto with arith.
  Qed.

  Proposition enum_P_prop4 : forall y, P y -> { x | y = enum_P x }.
  Proof.
    intros y Hy.
    destruct (le_lt_dec (enum_P 0) y) as [ H | H ].
    apply fnat_find in H.
    destruct H as [ x [ H1 H2 ] ].
    destruct (le_lt_dec y (enum_P x)) as [ H3 | H3 ].
    exists x; auto with arith.
    apply (enum_P_prop3 x y) in Hy; tauto.
    apply enum_P_sinc.
    apply enum_P_prop2 in H.
    contradiction (H Hy).
  Qed.

  Theorem enum_P_img : P ~ d_image enum_P.
  Proof.
    split.
    intros y Hy; destruct (enum_P_prop4 _ Hy) as [ x ]; exists x; auto.
    intros y [ x Hx ]; rewrite <- Hx; apply enum_P_prop1.
  Qed.
  
End enum_P.

Section NatDecPred2.

  Variable P : nat -> nat -> Prop.
  Hypothesis Pdec : forall a b, { P a b } + { ~ P a b }.
  Hypothesis Prefl : forall a, P a a.

  Definition pnat2_minimize x : { m | P x m /\ forall j, P x j -> m <= j }.
  Proof.
    apply (pnat_minimize_a (P x)) with (a := x); auto.
    intro; auto.
  Qed.

End NatDecPred2.

Section NatQuotient.

  (* compute the increasing enumeration of an infinite set of classes of nat *)

  Variable eq : nat -> nat -> Prop.
  Variable eq_dec : forall x y, { eq x y } + { ~ eq x y }.
  Variable eq_refl : forall x, eq x x.
  Variable eq_sym : forall x y, eq x y -> eq y x.
  Variable eq_tran : forall x y z, eq x y -> eq y z -> eq x z.

  (* f is an injective map from nats to eq_classes 
      ensuring that there are infinitely many classes 
  *)
 
  Variable f : nat -> nat.
  Hypothesis f_inj : forall x y, eq (f x) (f y) -> x = y.

  Let class_sig := pnat2_minimize _ eq_dec eq_refl.

  Let class x := proj1_sig (class_sig x).

  Let class_bij x y : class x = class y <-> eq x y.
  Proof.
    unfold class.
    generalize (proj2_sig (class_sig x)) (proj2_sig (class_sig y)).
    intros [ Ha1 Ha2 ] [ Hb1 Hb2 ]; split. 
    intros Hab.
    apply eq_tran with (1 := Ha1), eq_sym, eq_tran with (1 := Hb1), eq_sym.
    rewrite Hab; auto.
    intros  Hxy.
    apply le_antisym.
    apply Ha2, eq_tran with (1 := Hxy); trivial.
    apply Hb2, eq_sym, eq_tran with (2 := Hxy), eq_sym; trivial.
  Qed.

  Let class_eq x : eq x (class x).
  Proof. unfold class; generalize (proj2_sig (class_sig x)); intuition. Qed.

  Let class_decr x : class x <= x.
  Proof. unfold class; generalize (proj2_sig (class_sig x)); intuition. Qed.

  Let class_class x : class (class x) = class x.
  Proof. apply class_bij, eq_sym; auto. Qed.

  (* class x is the minimal representant of the class of x *)

  Let g x := class (f x).
  Let g_inj : fnat_inj g.
  Proof. unfold g; intros x y Hxy; apply f_inj, class_bij; trivial. Qed.

  Let g_inf := fnat_inj_prop _ g_inj.

  Let P x := class x = x.
  Let Pdec : pnat_dec P.
  Proof. intros x; unfold P; apply eq_nat_dec. Qed.

  Let Pinf : pnat_inf P.
  Proof.
    intros x.
    destruct (g_inf x) as [ y Hy ].
    exists (g y).
    split; (omega || trivial).
    unfold g, P; auto.
  Qed.

  Definition quot_repr := enum_P _ Pdec Pinf.
  Let quot_repr_sinc : fnat_sinc quot_repr.
  Proof. apply enum_P_sinc. Qed.

  Definition quot_class x := fnat_sinc_inv _ quot_repr_sinc (class x).
  
  Theorem quot_class_repr x : quot_class (quot_repr x) = x.
  Proof.
    unfold quot_class, quot_repr.
    generalize (enum_P_prop1 _ Pdec Pinf x); intros Hx; red in Hx. 
    rewrite Hx; apply fnat_sinc_inv_prop.
  Qed.

  Theorem quot_class_fun x y : eq x y -> quot_class x = quot_class y.
  Proof. intros Hxy; unfold quot_class; rewrite (proj2 (class_bij _ _) Hxy); trivial. Qed.

  Theorem quot_repr_class x : eq (quot_repr (quot_class x)) x.
  Proof.
    unfold quot_repr.
    generalize (enum_P_prop1 _ Pdec Pinf (quot_class x)).
    intros Hx. red in Hx.
    apply class_bij.
    rewrite Hx.
    unfold quot_class.
    destruct (enum_P_prop4 _ Pdec Pinf _ (class_class x)) as [ y Hy ].
    rewrite Hy.
    rewrite fnat_sinc_inv_prop; trivial.
  Qed.

  Variable h : nat -> nat.
  Hypothesis Hh : forall x y, eq x y -> h x = h y.
  
  Definition quot_lift : { t : nat -> nat | forall n x, eq (quot_repr n) x -> t n = h x }.
  Proof.
    exists (fun n => h (quot_repr n)).
    intros n x Hx; apply Hh; auto.
  Qed.

End NatQuotient.

(* We give a constructive proof of weaker version of the Cantor-Bernstein Theorem for the type nat

   any type T with a surjection nat -> T and strictly size increasing sequence within it
   can be indexed by nat 
*) 

Section Indexation.

  Variable T : Type.
  Variable dec : nat -> T.
  Variable enc : T -> nat.
  Hypothesis encdec_surj : forall t, dec (enc t) = t.

  Let enc_inj t1 t2 : enc t1 = enc t2 -> t1 = t2.
  Proof.
    intros H.
    rewrite <- encdec_surj, <- H, encdec_surj.
    trivial.
  Qed.
    
  Variable seq : nat -> T.
  Variable sze : T -> nat.
  Hypothesis seq_inc : forall n, sze (seq n) < sze (seq (S n)).

  Let eq i j := enc (dec i) = enc (dec j).

  Let eq_encdec n : eq (enc (dec n))  n.
  Proof. red. rewrite encdec_surj. trivial. Qed.

  Let eq_dec : forall a b, { eq a b } + { ~ eq a b }.
  Proof. intros; apply eq_nat_dec. Qed.

  Let eq_refl a : eq a a.                         Proof. red; auto. Qed.
  Let eq_sym a b : eq a b -> eq b a.              Proof. unfold eq; auto. Qed.
  Let eq_tran a b c : eq a b -> eq b c -> eq a c. Proof. unfold eq; intro H; rewrite H; auto. Qed.

  Let g n := sze (seq n).

  Let g_inj : fnat_inj g.
  Proof. apply fnat_sinc_imp_inj, fnat_inc_imp_sinc; auto. Qed.
    
  Let f n := enc (seq n).
  Let f_inj x y : eq (f x) (f y) -> x = y.
  Proof.
    unfold f, eq. 
    rewrite encdec_surj, encdec_surj.
    intros Hxy.
    unfold f in Hxy.
    apply enc_inj in Hxy.
    apply g_inj.
    unfold g.
    rewrite Hxy.
    trivial.
  Qed.

  Definition index_dec n := dec (quot_repr _ eq_dec eq_refl eq_sym eq_tran _ f_inj n).
  Definition index_enc t := quot_class _ eq_dec eq_refl eq_sym eq_tran _ f_inj (enc t).

  Theorem index_enc_dec n : index_enc (index_dec n) = n.
  Proof.
    unfold index_enc, index_dec.
    rewrite quot_class_fun with (y := quot_repr _ eq_dec eq_refl eq_sym eq_tran _ f_inj n); trivial.
    apply quot_class_repr.
  Qed.

  Theorem index_dec_enc t : index_dec (index_enc t) = t.
  Proof.
    unfold index_dec, index_enc.
    generalize (quot_repr_class  _ eq_dec eq_refl eq_sym eq_tran _ f_inj (enc t)).
    intros Ht. unfold eq in Ht |- *. 
    apply enc_inj in Ht.
    rewrite Ht.
    trivial.
  Qed.

End Indexation.



    

  
    
    