(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.

Require Import predicate.
Require Import maps.
Require Import cantorbernstein.

Record partial_enumeration {X} (P : X -> Prop) := MkEnum {
  enum_dmap : nat -> X;
  enum_rmap : X -> nat;
  enum_surj : forall x, P x -> enum_dmap (enum_rmap x) = x
}.

Proposition penum_id X (A B : X -> Prop) : A ~ B -> partial_enumeration A -> partial_enumeration B.
Proof.
  intros Id HA.
  destruct HA as [ dA rA sA ].
  exists dA rA.
  intros x Hx. apply sA.
  apply (proj2 Id). trivial.
Qed.

Definition enumeration X := partial_enumeration (fun _ : X => True).

Proposition enumeration_surj X (eX : enumeration X) : surjective (enum_dmap _ eX).
Proof.
  intros x. destruct eX as [dX rX sX]. simpl. 
  exists (rX x). rewrite sX; auto.
Qed.

Proposition enumeration_inj X (eX :enumeration X) : injective (enum_rmap _ eX).
Proof.
  intros x x' H. destruct eX as [dX rX sX].
  simpl in H. 
  rewrite <- (sX x), <- (sX x'), -> H; trivial.
Qed.

Proposition enumeration_dec X : enumeration X -> forall x y : X, { x = y } + { x <> y }.
Proof.
  intros [ dmap rmap surj ] x y.
  destruct (eq_nat_dec (rmap x) (rmap y)) as [ H | H ]; [ left | right ].
  rewrite <- surj, <- H; auto; symmetry; auto.
  intros Hx; apply H; rewrite Hx; auto.
Qed.
  
Record indexation X := MkIndex {
  idx_dmap : nat -> X;
  idx_rmap : X -> nat;
  idx_inj : forall x, idx_rmap (idx_dmap x) = x;
  idx_surj : forall x, idx_dmap (idx_rmap x) = x
}.

Definition enum_from_idx X : indexation X -> enumeration X.
Proof.
  intros HX. exists (idx_dmap _ HX) (idx_rmap _ HX).
  intros x _. apply idx_surj.
Qed.

Proposition indexation_dinj X (iX : indexation X) : injective (idx_dmap _ iX).
Proof.
  intros x1 x2 Hy.
  rewrite <- (idx_inj _ iX), <- Hy, (idx_inj _ iX).
  auto.
Qed.

Proposition indexation_dsurj X (iX : indexation X) : surjective (idx_dmap _ iX).
Proof.
  intros x.
  exists (idx_rmap _ iX x).
  destruct iX; auto.
Qed.

Proposition indexation_rinj X (iX : indexation X) : injective (idx_rmap _ iX).
Proof.
  intros x1 x2 Hy.
  rewrite <- (idx_surj _ iX), <- Hy, (idx_surj _ iX).
  auto.
Qed.

Proposition indexation_rsurj X (iX : indexation X) : surjective (idx_rmap _ iX).
Proof.
  intros x.
  exists (idx_dmap _ iX x).
  destruct iX; auto.
Qed.

Definition nat_index : indexation nat.
Proof.
  exists (fun x => x) (fun x => x);
  reflexivity.
Qed.

Record isequence X := MkISequence {
  iseq_seq : nat -> X;
  iseq_sze : X -> nat;
  iseq_inc : forall n, iseq_sze (iseq_seq n) < iseq_sze (iseq_seq (S n))
}.

(* for an indexation, there must be an infinite number of elements in X, 
    which is witnessed by a size-increasing of elements of X             
*)

Definition idx_from_enum_iseq X : enumeration X -> isequence X -> indexation X.
Proof.
  intros [ dmap rmap Hdr ] [ iseq isze iinc ].
  assert (forall x, dmap (rmap x) = x) as H.
  intros x; apply Hdr; trivial. clear Hdr.
  set (dm := index_enc _ _ _ H _ _ iinc).
  set (rm := index_dec _ _ _ H _ _ iinc).
  exists rm dm; [ apply index_enc_dec | apply index_dec_enc ].
Qed.


