(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Arith.
Require Import NPeano.

Require Import predicate.
Require Import carrier.
Require Import finite.

Section Definitions.

  Local Set Implicit Arguments.

  Variable X Y : Type.
  Definition graph := X * Y -> Prop.

  Variable f : graph.

  Definition functional_graph :=  forall x y y', f (x,y) -> f (x,y') -> y = y'.
  Definition total_graph := forall x, exists y, f (x,y).
  
  Definition gr_dimage P y := exists x, P x /\ f (x,y).
  Definition gr_rimage Q x := exists y, Q y /\ f (x,y).

  Definition gr_fst := gr_dimage (fun _ => True).
  Definition gr_snd := gr_rimage (fun _ => True).

End Definitions.

Section Graph.

  Variable X Y : Type.

  Implicit Type f g : graph X Y.
  Implicit Type A B : X -> Prop.
  Implicit Type C D : Y -> Prop.
  Implicit Type x : X.
  Implicit Type y : Y.

  Proposition gr_dimage_sg f x : gr_dimage f (sg x) ~ fun y => f (x,y).
  Proof.
    split; intro.
    intros [ ? [ H ] ].
    pred_solve; rewrite -> H; trivial.
    exists x; split; auto.
  Qed.

  Proposition gr_dimage_sg' f x y : gr_dimage f (sg x) ~ sg y -> f (x,y).
  Proof. intros [ _ H ]. pred_solve. apply (proj1 (gr_dimage_sg _ _) _ H). Qed. 

  Proposition gr_dimage_inc_src f A B : A inc B -> gr_dimage f A inc gr_dimage f B.
  Proof. intros ? ? [ x [ ? ? ] ]; exists x; auto. Qed.

  Hint Resolve gr_dimage_inc_src.

  Proposition gr_dimage_id_src f A B : A ~ B -> gr_dimage f A ~ gr_dimage f B.
  Proof. intros []; split; auto. Qed.

  Proposition gr_dimage_inc f g A : f inc g -> gr_dimage f A inc gr_dimage g A.
  Proof. intros ? ? [ x [ ] ]; exists x; split; auto. Qed.

  Hint Resolve gr_dimage_inc.

  Proposition gr_dimage_id f g K : f ~ g -> gr_dimage f K ~ gr_dimage g K.
  Proof. intros [ ]; split; auto. Qed.

  Proposition gr_dimage_empty_src f : gr_dimage f empty ~ empty.
  Proof.
    split; auto.
    intros ? ( ? & [] & ? ).
  Qed.

  Proposition gr_rimage_inc_dst f C D : C inc D -> gr_rimage f C inc gr_rimage f D.
  Proof. intros ? ? [ x [ ? ? ] ]; exists x; auto. Qed.

  Hint Resolve gr_rimage_inc_dst.

  Proposition gr_rimage_inc f g C : f inc g -> gr_rimage f C inc gr_rimage g C.
  Proof. intros ? ? [ x [ ] ]; exists x; split; auto. Qed.

  Hint Resolve gr_rimage_inc.

  Proposition gr_rimage_id f g C : f ~ g -> (gr_rimage f) C ~ (gr_rimage g) C.
  Proof. intros [ ]; split; auto. Qed.

  Hint Resolve gr_dimage_id gr_rimage_id.

  Proposition gr_dimage_empty A : gr_dimage empty A ~ (Emptyset Y).
  Proof. split; auto; intros ? [ ? [ _ H ] ]; contradict H. Qed.

  Proposition gr_rimage_empty C : gr_rimage empty C ~ (Emptyset X).
  Proof. split; auto; intros ? [ ? [ _ H ] ]; contradict H. Qed.

  Proposition gr_dimage_sg_gr x y A : A x -> gr_dimage (sg (x,y)) A ~ sg y.
  Proof.
    intros H; split.
    intros y' [ x' [ H1 H2 ] ]; pred_solve.
    injection H2; auto.
    intros y' Hy'; pred_solve.
    rewrite Hy'.
    exists x; auto.
  Qed.

  Proposition gr_fst_inc f g : f inc g -> gr_fst f inc gr_fst g.       Proof. unfold gr_fst; auto. Qed.
  Proposition gr_snd_inc f g : f inc g -> gr_snd f inc gr_snd g.       Proof. unfold gr_snd; auto. Qed.
  Proposition gr_fst_id f g : f ~ g -> gr_fst f ~ gr_fst g.            Proof. unfold gr_fst; auto. Qed.
  Proposition gr_snd_id f g : f ~ g -> gr_snd f ~ gr_snd g.            Proof. unfold gr_snd; auto. Qed.
 
  Proposition gr_fst_sg x y : gr_fst (sg (x,y)) ~ sg y.
  Proof.
    split; intros z Hz.    
    destruct Hz as [ k [ _ Hk ] ]; pred_solve; injection Hk; auto.
    pred_solve; exists x; split; auto; rewrite Hz; auto.
  Qed.

  Proposition gr_dimage_cup1 f g A : gr_dimage (f cup g) A ~ gr_dimage f A cup gr_dimage g A.
  Proof. 
    split; intros y Hy. 
    destruct Hy as [ x [ HK Hx ] ]. 
    pred_solve; [ apply bin_union_inc_r1 | apply bin_union_inc_r2 ]; 
      exists x; split; trivial.
    pred_solve; destruct Hy as [ x [ HK Hx ] ]; exists x;
      split; trivial; [ apply bin_union_inc_r1 | apply bin_union_inc_r2 ]; trivial.
  Qed.

  Proposition gr_rimage_cup1 f g C : gr_rimage (f cup g) C ~ gr_rimage f C cup gr_rimage g C.
  Proof. 
    split; intros y Hy. 
    destruct Hy as [ x [ HK Hx ] ]. 
    pred_solve; [ apply bin_union_inc_r1 | apply bin_union_inc_r2 ]; 
      exists x; split; trivial.
    pred_solve; destruct Hy as [ x [ HK Hx ] ]; exists x;
      split; trivial; [ apply bin_union_inc_r1 | apply bin_union_inc_r2 ]; trivial.
  Qed.

  Proposition gr_dimage_cup2 f A B : gr_dimage f (A cup B) ~ gr_dimage f A cup gr_dimage f B.
  Proof.
    split. 
    intros y (x & HK & Hf).
    pred_solve; [ apply bin_union_inc_r1 | apply bin_union_inc_r2 ]; exists x; tauto.
    pred_solve.
  Qed.
    
  Proposition gr_rimage_cup2 f C D : gr_rimage f (C cup D) ~ gr_rimage f C cup gr_rimage f D.
  Proof.
    split. 
    intros y (x & HK & Hf).
    pred_solve; [ apply bin_union_inc_r1 | apply bin_union_inc_r2 ]; exists x; tauto.
    pred_solve.
  Qed.

  Hint Immediate gr_fst_sg gr_dimage_cup1 gr_dimage_cup2 gr_rimage_cup1 gr_rimage_cup2.
  Hint Resolve gr_fst_inc.
 
  Proposition gr_fst_cup f g : gr_fst (f cup g) ~ gr_fst f cup gr_fst g.    Proof. unfold gr_fst; auto. Qed. 
  Proposition gr_snd_cup f g : gr_snd (f cup g) ~ gr_snd f cup gr_snd g.    Proof. unfold gr_snd; auto. Qed. 

  Let gr_fst_finite_rec f l : carrier l inc gr_fst f -> exists l', carrier l' inc f /\ carrier l ~ gr_fst (carrier l').
  Proof.
    induction l as [ | y l ]; intros Hl. 
    exists nil. split; auto. split; auto. 
    intros ? [ ? [ _ Hx ] ]; contradiction Hx.
    apply carrier_cons_1 in Hl; destruct Hl as [ Hy Hl ].
    destruct IHl as (l' & Hl' & H''); auto. 
    destruct Hy as [ x [ _ Hx ] ].
    exists ((x,y)::l'); split; auto.
 
    split. 

    apply carrier_cons_2. 
    exists x; auto. 
    apply inc_transitive with (1 := proj1 H''); auto.

    intros y' [ x' [ _ Hx' ] ]; simpl in Hx'.
    destruct Hx' as [ Hx' | Hx' ].
    injection Hx'; intros H1 _; rewrite -> H1; auto. 
    right; apply H''; exists x'; auto.
  Qed.

  Proposition gr_fst_finite f Yf : finite Yf -> Yf inc gr_fst f -> exists hf, finite hf /\ hf inc f /\ Yf ~ gr_fst hf.
  Proof.
    intros [ l Hl ] H.
    destruct (gr_fst_finite_rec f l) as [ l' Hl' ].
    apply inc_transitive with (2 := H), Hl; auto.
    exists (carrier l'). 
    split. apply finite_carrier.
    split. tauto. 
    apply id_transitive with (1 := id_symetric _ _ _ Hl).
    tauto.
  Qed.
 
  Definition function_graph f := forall x, exists y, gr_dimage f (sg x) ~ sg y.

  Proposition function_graph_functional f : function_graph f -> functional_graph f.
  Proof.
    intros Hf x y y' H H'.
    destruct (Hf x) as [ k Hfx ].
    apply id_symetric in Hfx.
    apply (proj2 (id_transitive _ _ _ _ Hfx (gr_dimage_sg _ _))) in H.
    apply (proj2 (id_transitive _ _ _ _ Hfx (gr_dimage_sg _ _))) in H'.
    pred_solve; rewrite <- H; trivial.
  Qed.

  Proposition function_graph_total f : function_graph f -> total_graph f.
  Proof.
    intros Hf x.
    destruct (Hf x) as [ k Hfx ].
    exists k; apply gr_dimage_sg' in Hfx; auto.
  Qed.

  Proposition functional_total_graph_function f : functional_graph f -> total_graph f -> function_graph f.
  Proof.
    intros H1 H2 x.
    destruct (H2 x) as ( y & Hy ).
    exists y.
    split; pred_solve.
    intros z (q & H3 & H4).
    pred_solve; rewrite H3 in Hy.
    rewrite (H1 _ _ _ Hy H4); auto.
    exists x; auto.
  Qed.
    
  Theorem function_compose_function_graph f h : function_graph f -> function_graph (fun p => match p with (n,x) => f (h n, x) end).
  Proof.
    intros Hf x.
    destruct (Hf (h x)) as [ y Hy ].
    exists y.
    apply gr_dimage_sg' in Hy.
    split.
    intros z ( k & Hk1 & Hk2 ).
    pred_solve.
    rewrite Hk1 in Hy.
    rewrite (function_graph_functional _ Hf _ _ _ Hy Hk2); auto.
    pred_solve.
    exists x; auto.
  Qed.

End Graph.

Implicit Arguments function_graph [X Y].

Hint Resolve gr_dimage_inc.
Hint Resolve gr_rimage_inc.
Hint Resolve gr_dimage_id gr_rimage_id.
Hint Resolve gr_dimage_inc_src gr_rimage_inc_dst.
Hint Immediate gr_dimage_sg gr_fst_sg gr_dimage_cup1 gr_dimage_cup2 gr_rimage_cup1 gr_rimage_cup2.

Proposition gr_dimage_sect {X} n (f : graph nat X) x : functional_graph f -> f (n,x) -> gr_dimage f (fun i => i < S n) ~ gr_dimage f (fun i => i < n) cup sg x.
Proof.
  intros H1 H2.
  split.
  intros y ( z & Hz1 & Hz2 ).
  destruct (eq_nat_dec z n) as [ Hz | Hz ].
  rewrite Hz in Hz2.
  apply bin_union_inc_r2.
  rewrite (H1 _ _ _ H2 Hz2); auto.
  apply bin_union_inc_r1.
  exists z.
  split; auto.
  apply lt_n_Sm_le, le_lt_or_eq in Hz1.
  destruct Hz1 as [ | Hz1 ]; auto.
  contradiction (Hz Hz1).
  pred_solve.
  apply gr_dimage_inc_src.
  intros y Hy. apply lt_trans with (1 := Hy); auto.
  exists n; auto.
Qed.

Section Bounded_graph.

  Variable X : Type.

  Definition bounded_graph n (f : graph nat X) := (forall k x, f (k,x) -> k<n) /\ (forall k, k<n -> exists x, gr_dimage f (sg k) ~ sg x). 

  Proposition bounded_graph_id n P Q : P ~ Q -> bounded_graph n P -> bounded_graph n Q.
  Proof.
    intros [ E1 E2 ] [ H1 H2 ].
    split.
    intros ? x ?; apply (H1 _ x); auto.
    intros k H; apply H2 in H; destruct H as [ x H ]; exists x.
    apply id_transitive with (2 := H); auto.
  Qed.

  Proposition bounded_graph_functional n f : bounded_graph n f -> functional_graph f.
  Proof.
    intros [ H1 H2 ] x y y' H3 H4.
    generalize (H1 _ _ H3); intros H5.
    generalize (H2 _ H5); intros [ z H6 ].
    apply id_symetric in H6.
    generalize (id_transitive _ _ _ _ H6 (gr_dimage_sg _ _ f x)).
    intros H7. 
    apply H7, in_sg_eq in H3.
    apply H7, in_sg_eq in H4.
    rewrite <- H3; auto.
  Qed.

  Proposition gr_dimage_bounded n f K : bounded_graph n f -> gr_dimage f K inc gr_dimage f (K cap (fun i => i < n)).
  Proof.
    intros Hf y [ x [ H1 H2 ] ].
    exists x. split; auto.
    apply bin_inter_sg'; auto.
    apply (proj1 Hf _ _ H2).
  Qed.

  Proposition bounded_graph_fst n f : bounded_graph n f -> gr_fst f ~ gr_dimage f (fun i => i < n).
  Proof.
    intros Hf; split;
    unfold gr_fst.
    apply inc_transitive with (1 := gr_dimage_bounded _ _ _ Hf).
    apply gr_dimage_inc_src; auto.
    apply gr_dimage_inc_src; intro; auto.
  Qed.
 
  Proposition bounded_graph_O : bounded_graph 0 empty.
  Proof.
    split.
    intros ? ? H; contradiction H.
    intros ? H; contradiction (le_Sn_O _ H).
  Qed.

  Proposition bounded_graph_S n P x : bounded_graph n P -> bounded_graph (S n) (P cup sg (n,x)).
  Proof.
    intros [ H0 H1 ].
    split.

    intros k y H'. apply bin_union_sg in H'. destruct H' as [ H' | H' ].
    apply (lt_trans _ n); auto; apply (H0 _ _ H'). 
    apply in_sg_eq in H'; injection H'; intros _ H''; rewrite -> H''; auto. 

    intros k Hk.
    destruct (le_lt_dec n k) as [ H | H ].
    apply le_S_n in Hk. rewrite -> (le_antisym _ _ Hk H).
    clear k Hk H.
    exists x. split; intros y Hy. apply (proj1 (gr_dimage_sg _ _ _ _)) in Hy.
    apply bin_union_sg in Hy. destruct Hy as [ Hy | Hy ].
    apply H0 in Hy. contradiction (lt_irrefl _ Hy).
    apply in_sg_eq in Hy. injection Hy. intros Eq. rewrite -> Eq; auto. 
    apply in_sg_eq in Hy. rewrite -> Hy. apply (proj2 (gr_dimage_sg _ _ _ _)). 
    apply bin_union_inc_r2; auto.

    destruct (H1 _ H) as [ y Hy ]. exists y.
    apply id_transitive with (2 := Hy).
    split; auto.
    intros z Hz. apply (proj1 (gr_dimage_sg _ _ _ _)) in Hz.
    apply bin_union_sg in Hz. destruct Hz as [ Hz | Hz ].
    exists k; auto.
    apply in_sg_eq in Hz.
    contradict H; injection Hz; intros _ H3; rewrite H3; apply lt_irrefl.
  Qed.

  Section Restriction.
  
    Variable n : nat.
    Variable f : graph nat X.

    Definition restriction p := match p with (k,c) => k < n /\ f p end.

    Proposition restriction_dimage_inc Y : gr_dimage restriction Y inc gr_dimage f Y.
    Proof.
      intros x ( y & Hy1 & Hy2 & Hy3 ).
      exists y; auto.
    Qed.

    Hypothesis function_f : function_graph f.

    Proposition restriction_bounded : bounded_graph n restriction.
    Proof.
      split. intros ? _ (?,_); auto.
      intros k H. destruct (function_f k) as [ x Hx ].
      exists x.
      apply id_transitive with (2 := Hx).
      split.
      intros y [ q [ H1 [ H2 H3 ]] ]. exists q; auto.
      intros y [ q [ H1 H2 ] ].
      exists q; split; auto.
      split; auto.
      apply in_sg_eq in H1.
      rewrite <- H1; auto.
    Qed.

    Proposition restriction_dimage : gr_fst restriction ~ (gr_dimage f) (fun k => k < n).
    Proof.
      split; intros k Hk. destruct Hk as [ y [ _ Hy ] ]. red in Hy.
      exists y. trivial.
      destruct Hk as [ y [ Hx Hy ] ]. exists y. split; trivial.
      split; trivial.
    Qed.

  End Restriction.

  Section Restriction'.
 
    Proposition restriction_bounded' n m f : n <= m -> bounded_graph m f -> bounded_graph n (restriction n f).
    Proof.
      intros Hnm Hf. 
      generalize (bounded_graph_functional _ _ Hf); intros Hf0.
      destruct Hf as (Hf1 & Hf2).
      split.
      intros ? ? [ ]; auto.
      intros k Hk.
      destruct (Hf2 k) as [ x Hx ].
      apply lt_le_trans with (2 := Hnm); auto.
      exists x.
      apply gr_dimage_sg' in Hx.
      split.
      intros y (z & Hz1 & Hz2 & Hz3).
      apply in_sg_eq in Hz1.
      rewrite <- Hz1 in Hz3.
      rewrite (Hf0 _ _ _ Hx Hz3).
      auto.
      apply sg_inc.
      exists k.
      repeat (split; auto).
    Qed.

    Proposition restriction_dimage' n f Y : Y inc (fun i => i < n) -> gr_dimage f Y ~ gr_dimage (restriction n f) Y.
    Proof.
      intros HY.
      split.
      intros x ( y & Hy1 & Hy2 ).
      exists y.
      split; auto.
      apply HY in Hy1.
      split; auto.
      intros x ( y & Hy1 & Hy2 & Hy3 ).
      exists y; auto.
    Qed.

  End Restriction'.

  Section Extension.

    Variable n : nat.
    Variable f : graph nat X.
    Hypothesis bounded_f : bounded_graph n f.

    Definition extension c0 p :=
     match p with (k,c) => match (le_lt_dec n k) with right _ => f (k,c) | left _ => c = c0  end end.

    Proposition extension_function c0 : function_graph (extension c0).
    Proof.
      destruct bounded_f as [ Hf1 Hf2 ].
      intros x.
      case_eq (le_lt_dec n x); intros Hn H.
      exists c0.
      apply id_transitive with (1 := gr_dimage_sg _ _ _ _). 
      unfold extension; rewrite H.
      split; intros q H2. rewrite H2; auto. apply in_sg_eq in H2; auto.
      destruct (Hf2 _ Hn) as [ z Hz ].
      exists z. 
      apply id_transitive with (1 := gr_dimage_sg _ _ _ _).
      unfold extension; rewrite H.
      apply id_transitive with (2 := Hz).
      apply id_symetric; auto.
    Qed.

    Proposition extension_dimage1 c0 (K : nat -> Prop) j : n <= j -> K j -> gr_dimage (extension c0) K ~ gr_dimage f K cup sg c0.
    Proof.
      intros Hj HK.
      split.
      intros z Hz. destruct Hz as [ x [ Hx Hz ] ]. unfold extension in Hz.
      case_eq (le_lt_dec n x); intros H' H''; rewrite H'' in Hz.
      apply bin_union_inc_r2; rewrite -> Hz; auto. 
      apply bin_union_inc_r1; exists x; auto. 
      intros z Hz. apply bin_union_sg in Hz. red. 
      destruct Hz as [ [ x [ Hx Hz ] ] | Hz ].
      exists x. split; trivial. red.
      destruct (le_lt_dec n x) as [ H | ]; auto.
      apply (proj1 bounded_f) in Hz.
      contradict (lt_irrefl _ (lt_le_trans _ _ _ Hz H)).
      exists j. split; auto.
      red. destruct (le_lt_dec n j) as [ _ | H' ]. apply in_sg_eq in Hz; auto.
      contradiction (lt_irrefl _ (le_lt_trans _ _ _ Hj H')).
    Qed.

    Proposition extension_dimage2 c0 (K : nat -> Prop) : K inc (fun j => j < n) -> gr_dimage (extension c0) K ~ gr_dimage f K.
    Proof.
      intros HK.
      split.
      intros z Hz. destruct Hz as [ x [ Hx Hz ] ]. unfold extension in Hz.
      destruct (le_lt_dec n x) as [ H' | H' ]. apply HK in Hx.
      contradiction (lt_irrefl _ (le_lt_trans _ _ _ H' Hx)).
      exists x; split; trivial.
      intros z Hz. red. destruct Hz as [ x [ Hx Hz ] ].
      exists x. split; trivial. apply HK in Hx.
      red. destruct (le_lt_dec n x) as [ H | _ ].
      contradiction (lt_irrefl _ (le_lt_trans _ _ _ H (proj1 bounded_f _ _ Hz))).
      trivial.
    Qed.

    Proposition extension_gr_fst c0 : gr_fst (extension c0) ~ (gr_fst f) cup sg c0.
    Proof. apply (extension_dimage1 c0 _ n); auto. Qed.

  End Extension.

  Variable f : nat -> graph nat X.
  Variable i : nat -> nat.
  Hypothesis i_sup_id : forall n, n <= i n.
  Hypothesis f_bounded : forall n, bounded_graph (i n) (f n).
  Hypothesis f_inc : forall n m, n <= m -> f n inc f m.

  Let f_uniq n1 n2 k x y : f n1 (k,x) -> f n2 (k,y) -> x = y.
  Proof.
    intros H1 H2.
    set (m := max k (max n1 n2)).
    assert (k < S m) as H0'. 
      apply le_n_S, Nat.le_max_l.
    assert (n1 <= S m) as H1'.
      unfold m.
      apply le_trans with (2 := le_n_Sn _).
      apply le_trans with (2 := Nat.le_max_r _ _).
      apply Nat.le_max_l.
    assert (n2 <= S m) as H2'. 
      unfold m.
      apply le_trans with (2 := le_n_Sn _).
      apply le_trans with (2 := Nat.le_max_r _ _).
      apply Nat.le_max_r.
    apply f_inc with (1 := H1') in H1.
    apply f_inc with (1 := H2') in H2.
    apply (bounded_graph_functional _ _ (f_bounded (S m)) _ _ _ H1 H2).
  Qed.

  Let f_union_uniq_rec k x : union f (k,x) -> exists n, f n (k,x).
  Proof. apply union_sg. Qed.

  Let f_union_uniq k x y : union f (k,x) -> union f (k,y) -> x = y.
  Proof.
    intros Hx Hy.
    apply f_union_uniq_rec in Hx.
    apply f_union_uniq_rec in Hy.
    destruct Hx as [ n1 Hx ].
    destruct Hy as [ n2 Hy ].
    apply (f_uniq _ _ _ _ _ Hx Hy).
  Qed.

  Let f_union_exists k : exists x, union f (k,x).
  Proof.
    destruct (f_bounded (S k)) as [ H0 H1 ]. 
    assert (k < i (S k)) as H. 
      apply lt_le_trans with (1 := lt_n_Sn _); auto. 
    apply H1 in H.
    destruct H as [ x H ].
    exists x; exists (S k).
    apply proj2 in H.
    apply in_sg in H.
    apply (proj1 (gr_dimage_sg _ _ _ _)) in H; auto.
  Qed.

  Proposition bounded_inc_union : function_graph (union f).
  Proof.
    intros n.
    destruct (f_union_exists n) as [ x Hx ].
    exists x. split.
    intros y Hy.
    apply (proj1 (gr_dimage_sg _ _ _ _)) in Hy.
    rewrite -> (f_union_uniq _ _ _ Hx Hy); auto.
    apply sg_inc.
    apply (proj2 (gr_dimage_sg _ _ _ _)); auto.
  Qed.

  Proposition restriction_union n : restriction (i n) (union f) ~ f n.
  Proof.
    split; intros (k,x) Hx.
    destruct Hx as [ Hk Hx ].
    apply f_union_uniq_rec in Hx.
    destruct Hx as [ k' Hk' ].
    destruct (proj2 (f_bounded n) _ Hk) as [ x' [ _ Hx' ] ].
    apply in_sg in Hx'. 
    apply (proj1 (gr_dimage_sg _ _ _ _)) in Hx'.
    rewrite -> (f_uniq _ _ _ _ _ Hk' Hx'); auto.
    split. 
    apply (proj1 (f_bounded n) _ x); auto.
    exists n; auto.
  Qed.
 
End Bounded_graph.

Implicit Arguments bounded_graph [X].
Implicit Arguments restriction [X].

Hint Immediate bounded_graph_O gr_dimage_empty gr_rimage_empty.
Hint Resolve bounded_graph_S gr_dimage_inc_src.

Section Bounded_graph'.

  Variable X : Type.
  Variable f : nat -> graph nat X.
  Variable n0 : nat.
  Hypothesis f_bounded : forall n, bounded_graph (n0+n) (f n).
  Hypothesis f_inc : forall n m, n <= m -> f n inc f m.

  Let n0n n : n <= n0+n.
  Proof.
    rewrite -> plus_comm. apply le_plus_trans; auto.
  Qed.

  Proposition bounded_inc_union_n0 : function_graph (union f).
  Proof. apply (bounded_inc_union _ _ (fun n => n0+n)); auto. Qed.

  Proposition restriction_union_n0 n : restriction (n0+n) (union f) ~ f n.
  Proof. apply restriction_union; trivial. Qed.

End Bounded_graph'.

