(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import list_in_t.
Require Import list_fall_exst.
Require Import list_fall_exst_t.

Require Import predicate.
Require Import carrier.
Require Import finite.

Require Import system.

Section Proof_compositions.

  Variable stm : Type.

  Implicit Type h : stm.
  Implicit Type hh : list stm.

  Let proof_comp_1 s1 s2 h : proof s1 h -> forall c, proof (MkUAxiom h::s2) c -> proof (s1++s2) c.
  Proof.
    intros Hh.
    apply proof_rect.
    intros r hh c Hr I H.
    apply In_t_cons in Hr.
    destruct Hr as [ Hr | Hr ].

    rewrite Hr in I.
    red_instance I p I1 I2 I3.
    rewrite I2.
    apply (proof_incl (sys := s1)); auto.
    apply (in_proof r (hh := hh)); auto.
  Qed.
    
  Let proof_comp_2 s1 hh : Forall_t (proof s1) hh -> forall s2 c, proof (lstm_to_axioms hh++s2) c -> proof (s1++s2) c.
  Proof.
    induction hh; intros Hhh s2 c.
    simpl; apply proof_incl; auto.
    apply Forall_t_inv in Hhh;
    destruct Hhh as [ H1 H2 ].
    intros H3.
    simpl in H3.
    apply proof_comp_1 with (s1 := s1) in H3; auto.
    apply (proof_incl (sys' := lstm_to_axioms hh++s1++s2)) in H3.
    apply IHhh in H3; auto.
    apply proof_incl with (2 := H3); auto.
    apply In_t_inc_app; auto.
  Qed.

  Variable sys : list (rule stm). 

  Section compo_transfo. 

    Variable sys' : list (rule stm).

    Theorem proof_composition {hh} {c} : Forall_t (proof sys) hh -> proof (lstm_to_axioms hh++sys') c -> proof (sys++sys') c.
    Proof. intros; apply proof_comp_2 with (hh := hh); auto. Qed.

    Hypothesis Hsys : forall r hh c, In_t r sys -> instance r hh c -> proof (lstm_to_axioms hh++sys') c.
  
    Theorem proof_transformation : forall c, proof sys c -> proof sys' c.
    Proof.
      apply proof_rect.
      intros r hh c Hr I H.
      specialize (Hsys _ _ _ Hr I).
      apply proof_comp_2 with (s1 := sys') in Hsys; auto.
      apply proof_incl with (2 := Hsys); auto.
    Qed.

  End compo_transfo.

  Section compactness.

    Variable Ax : stm -> Prop.

    Theorem proof_compactness : forall c, proof_from sys Ax c -> { ax : list stm & { p : proof_with sys ax c | Forall Ax ax } }.
    Proof.
      apply proof_rect.
      intros r hh c Hr I H.
      apply In_t_cons in Hr.
      destruct Hr as [ Hr | Hr ].

      rewrite Hr in I.
      unfold MkAxioms in I.
      red_instance I c' I0 I1 I2.
      rewrite <- I1 in * |- *; clear c' I1 I2.
      exists (c::nil).
      assert (proof (lstm_to_axioms (c::nil)++sys) c) as Pr.
      apply (in_proof (sys := lstm_to_axioms (c::nil)++sys) (MkUAxiom c) (hh := nil)); simpl; auto.
      prove_instance tt; simpl; auto.
      exists Pr; auto.

      assert ({ hax : list stm  & { ll : Forall_t (proof (lstm_to_axioms hax++sys)) hh | Forall Ax hax } }) as H'.
      clear I. induction hh.
      exists nil.
      exists (in_Forall_t_0 _); auto.
      apply Forall_t_inv in H.
      destruct H as [ Ha H ].
      apply IHhh in H; clear IHhh.
      destruct H as [ Haxl [ Prl H ] ].
      destruct Ha as [ Hax [ Pr H' ] ].
      exists (Hax++Haxl).
      set(s2 := lstm_to_axioms (Hax++Haxl)++sys).
      assert(Forall_t (proof s2) (a::hh)) as Prl'.
      constructor.
      unfold s2. apply proof_incl with (2 := Pr).
      unfold lstm_to_axioms; rewrite map_app; auto.
      apply Forall_t_map with (2 := Prl).
      intros x. unfold s2. apply proof_incl.
      unfold lstm_to_axioms; rewrite map_app; auto.
      exists Prl'; auto.
      destruct H' as [ hax [ Hax H' ] ].
      exists hax.
      assert (In_t r (lstm_to_axioms hax++sys)) as Hr'; auto.
      exists (in_proof r Hr' I Hax); auto.
    Qed.

    Theorem closure_under_proof_compactness c : closure_under_proof sys Ax c -> exists axf, axf inc Ax /\ finite axf /\ closure_under_proof sys axf c.
    Proof.
      intros [ H ].
      apply proof_compactness in H.
      destruct H as [ axf [ H1 H2 ] ].
      exists (carrier axf).
      split; auto.
      intros ?; unfold carrier; apply Forall_forall; auto.
      split; auto.
      constructor.
      apply proof_with_from; auto.
    Qed.

  End compactness.

End Proof_compositions.

Implicit Arguments proof_composition [ stm sys sys' hh c ].
Implicit Arguments proof_transformation [ stm sys sys' ].
Implicit Arguments proof_compactness [ stm sys Ax c ].

Theorem proof_with_inst {stm} {sys} r {ax} {c : stm} : In_t r sys -> instance r ax c -> proof_with sys ax c.
Proof.
  intros Hr I. 
  assert (In_t r (lstm_to_axioms ax++sys)) as Hr'.
    apply In_t_app_r; auto.
  apply (in_proof r Hr' I).
  clear r Hr I Hr'.
  induction ax; constructor.
  apply (proof_0 (MkUAxiom a)); simpl; auto.
  exists tt; simpl; auto.
  assert (In_t_inc (lstm_to_axioms ax++sys) (lstm_to_axioms (a::ax)++sys)) as H.
    simpl; auto.
  apply (Forall_t_map _ (proof_incl H)) with (1 := IHax).
Qed.

Section simulation.

  Variable stm stm' : Type.
  Variable R : stm -> stm' -> Prop. 

  Variable sys  : list (rule stm).
  Variable sys' : list (rule stm').

  Hypothesis simul : forall c c', R c c' -> forall r hh, In_t r sys -> instance r hh c 
                    -> { hh' : list stm' & proof_with sys' hh' c' & Forall_R R hh hh' }. 

  Theorem proof_simulation c c' : R c c' -> proof sys c -> proof sys' c'.
  Proof.
    intros H1 H2; generalize c H2 c' H1; clear c c' H1 H2.
    induction 1 as [ r hh c Hr I Hhh ]; intros c' Hc'.
    apply (proof_incl (sys:=sys'++sys')); auto.
    destruct (simul _ _ Hc' _ _ Hr I) as [hh' H2 H1].
    apply proof_composition with (2 := H2).

    generalize hh Hhh H1; clear r I hh Hr Hhh H1 H2.
    induction hh'; intros hh Hhh H1.
    constructor.
    apply Forall_R_cons_r in H1;
    destruct H1 as (h & hh1 & Hh1 & Hh2 & Hh3).
    rewrite Hh2 in Hhh; apply Forall_t_inv in Hhh;
    destruct Hhh as [ Hh Hhh1 ].
    constructor.
    apply Hh; auto.
    apply IHhh' with (1 := Hhh1); auto.
  Qed.

End simulation.

Implicit Arguments proof_simulation [ stm stm' R sys sys' ].

Section simulation_rel_map.

  Variable stm stm' : Type.
  Variable f : stm -> stm'.  

  Variable R : rule stm -> rule stm' -> Prop.

  Hypothesis Hf : forall r r' hh c, R r r' -> instance r hh c -> instance r' (map f hh) (f c).

  Variable sys  : list (rule stm).
  Variable sys' : list (rule stm').

  Hypothesis HR : Forall_R R sys sys'.

  Theorem proof_simul_rel_map : forall c, proof sys c -> proof sys' (f c).
  Proof.
    induction 1 as [ r hh c Hr I Hhh ].
    destruct (Forall_R_In_t HR Hr) as [ r' H1 H2 ].
    apply in_proof with (r:=r') (hh:=map f hh); auto.
    apply (Hf r); auto.
    apply Forall_t_Forall_t_map.
    generalize Hhh.
    apply Forall_t_map.
    auto.
  Qed.

End simulation_rel_map.

Implicit Arguments proof_simul_rel_map [ stm stm' R sys sys' ].

Section projection.

  Variable stm stm' : Type.
  Variable f : stm -> stm'.

  Definition map_instance hh c := (map f hh, f c). 

  Variable sys : list (rule stm).
  Variable sys' : list (rule stm').

  Hypothesis Hmap : forall r hh c, In_t r sys -> instance r hh c -> { r' : rule stm' & In_t r' sys' & instance r' (map f hh) (f c) }.

  Let Hmap' : forall r hh c, In_t r sys -> instance r hh c -> proof_with sys' (map f hh) (f c).
  Proof.
    intros r hh c Hr I.
    destruct (Hmap _ _ _ Hr I) as [ r' H1 H2 ].
    apply (proof_with_inst r'); auto.
  Qed.

  Theorem proof_projection c : proof sys c -> proof sys' (f c).
  Proof.
    set (R x y := y = f x).
    apply proof_simulation with (R := R); try (red; auto).
    unfold R; clear c R.
    intros c c' H. rewrite H; clear c' H.
    intros r hh Hr I.
    exists (map f hh).
    apply (Hmap' _ _ _ Hr); auto.
    clear r c Hr I.
    exists (map (fun c => (c,f c)) hh).
    split.
    rewrite map_map, map_id; auto. 
    split.
    rewrite map_map. apply map_ext. intro; auto.
    apply Forall_forall.
    intros (c,c').
    intros H.
    apply in_map_iff in H.
    destruct H as [ x [ H _ ] ].
    injection H; clear H.
    intros H1 H2; rewrite H2 in H1; auto.
  Qed.

End projection.

Implicit Arguments proof_projection [ stm stm' f sys sys' ].

Section simulation_bij.

  Variable stm stm' : Type.
  Variable gr : rule stm -> rule stm'.

  Variable f : stm' -> stm.
  Variable g : stm  -> stm'.

  Hypothesis Hgr_1 : forall r hh c, instance (gr r) hh c -> instance r (map f hh) (f c).
  Hypothesis Hgr_2 : forall r hh c, instance r hh c -> instance (gr r) (map g hh) (g c).

  Variable sys : list (rule stm).

  Theorem proof_simul_bij_1 : forall c, proof (map gr sys) c -> proof sys (f c).
  Proof.
    apply proof_simul_rel_map with (R := fun x y => x = gr y).
    intros r r' hh c HR I.
    rewrite HR in I; apply Hgr_1; auto.
    exists (map (fun r => (gr r,r)) sys).
    split; auto.
    rewrite map_map; apply map_ext; intro; auto.
    split; auto.
    rewrite map_map, <- map_id at 1. apply map_ext; intro; auto.
    apply Forall_forall. 
    intros [ r' r ] H.
    apply in_map_iff in H.
    destruct H as ( x & H1 & _ ).
    congruence.
  Qed.

  Theorem proof_simul_bij_2 : forall c, proof sys c -> proof (map gr sys) (g c).
  Proof.
    apply proof_simul_rel_map with (R := fun x y => y = gr x).
    intros r r' hh c HR I.
    rewrite HR; apply Hgr_2; auto.
    exists (map (fun r => (r,gr r)) sys).
    split; auto.
    rewrite map_map, <- map_id at 1. apply map_ext; intro; auto.
    split; auto.
    rewrite map_map; apply map_ext; intro; auto.
    apply Forall_forall. 
    intros [ r' r ] H.
    apply in_map_iff in H.
    destruct H as ( x & H1 & _ ).
    congruence.
  Qed.
 
End simulation_bij.

Section proof_with_condition.

  Variable stm : Type.
  Variable P : stm -> Prop.

  Let new_pred pt (cnd : pt -> Prop) (rinst : pt -> (list stm)*stm) p := 
    match (rinst p) with (hh,c) => cnd p /\ P c /\ Forall P hh end.

  Definition CndRule (r : rule stm) := MkRule (new_pred _ (rpred r) (rinst r)) (rinst r). 

  Variable sys : list (rule stm).

  Let sys' := map CndRule sys.

  Theorem proof_forget_condition : forall c, proof sys' c -> proof sys c.
  Proof.
    apply proof_rect. 
    intros r' hh' c' Hr I Hhh'.
    unfold sys' in Hr.
    apply In_t_map_rev in Hr.
    destruct Hr as [r H1  H2 ].
    rewrite H2 in I. clear H2.
    red_instance I p H2 H3 H4.
    red in H2.
    rewrite <- rule_hyps_conc_inst in H2.
    rewrite <- H3, <- H4 in H2.
    destruct H2 as (H2 & H5 & H6).
    apply in_proof with (r := r) (hh := hh'); auto.
    prove_instance p; auto.
  Qed.

  Hypothesis HP : forall r hh c, In_t r sys -> instance r hh c -> P c -> Forall P hh.

  Theorem proof_with_condition c : P c -> proof sys c -> proof sys' c.
  Proof.
    intros H1 H2; generalize c H2 H1; clear c H1 H2.
    induction 1 as [ r hh c Hr I Hhh]; intros Hc.
    apply (in_proof) with (r := CndRule r) (hh := hh).

    unfold sys'; apply In_t_map; auto.

    red_instance I p H1 H2 H3.
    apply in_instance with (r := CndRule r) (p:=p); simpl; auto.
    red. rewrite <- rule_hyps_conc_inst, <- H2, <- H3.
    split; auto. 
    split; auto.
    apply (HP _ _ c Hr); auto.
    prove_instance p; auto.

    apply (HP _ _ _ Hr I) in Hc.
    apply In_t_imp_Forall_t.
    intros x Hx.
    generalize (Forall_t_imp_In_t Hhh Hx).
    intros H; apply H.
    apply In_t_In in Hx.
    apply (list_fall_app_0 _ _ _ Hc _ Hx).
  Qed.

End proof_with_condition.

Implicit Arguments proof_with_condition [ stm P sys ].

Section stm_with_info.

  Variable stm : Type.

  Variable f : stm -> Type.

  Section new_rule.

    Variable r : rule stm.

    Let X := { p : rparam r & Forall_t f (rhyps r p) & f (rconc r p) }.
  
    Implicit Type pX : X.    

    Let new_cnd  pX := match pX with existT2 p Hh Hc => rpred r p end.

    Let new_hyps pX := match pX with existT2 p Hh Hc => Forall_t_lsigT (rhyps r p) Hh end.
    Let new_conc pX := match pX with existT2 p Hh Hc => existT _ (rconc r p) Hc end.

    Definition MkRule_info := MkRule new_cnd (fun pX => (new_hyps pX, new_conc pX)).
 
  End new_rule.

  Let inst_info r (p : rparam r) hh (Hhh :  Forall_t f hh) c (Hc : f c) : 
          rpred r p -> hh = rhyps r p -> c = rconc r p -> instance (MkRule_info r) (Forall_t_lsigT hh Hhh) (existT _ c Hc).
  Proof.
    intros H1 H2 H3.
    subst.
    exists (existT2 (fun x => Forall_t f (rhyps r x)) (fun x => f (rconc r x)) p Hhh Hc).
    split; simpl; auto.
  Qed.

  Fact instance_with_info_prop1 r hh c : instance r (map (projT1 (P:=f)) hh) (projT1 c) -> instance (MkRule_info r) hh c.
  Proof.
    intros I.
    red_instance I p H1 H3 H2.
    destruct c as [ c Hc ].
    rewrite <- (lsigT_sigfall_t_prop hh) in * |- *.
    destruct (lsigT_sigfall_t hh) as [ h Hh ].
    clear hh.
    simpl in H2 |- *.
    apply inst_info with (p := p); auto.
    rewrite <- H2. symmetry.
    apply Forall_t_lsigT_prop.
  Qed.

  Fact instance_with_info_prop2 r hh c : instance (MkRule_info r) hh c -> instance r (map (projT1 (P:=f)) hh) (projT1 c).
  Proof.
    intros I.
    red_instance I p H1 H3 H2.
    destruct p as [ p H4 H5 ].
    rewrite H2, H3. simpl.
    rewrite Forall_t_lsigT_prop.
    prove_instance p; auto.
  Qed.
  
  Variable sys : list (rule stm).

  Let sys' := map MkRule_info sys.

  Theorem proof_forget_info : forall c, proof sys' c -> proof sys (projT1 c).
  Proof.
    induction 1 as [ r' hh' c' Hr I Hhh' ].
    unfold sys' in Hr.
    apply In_t_map_rev in Hr.
    destruct Hr as [r H1  H2 ].
    rewrite H2 in I.
    apply instance_with_info_prop2 in I.
    apply in_proof with (2 := I); auto.
    apply Forall_t_Forall_t_map; auto.
  Qed.
    
  Hypothesis Hf : forall r hh c, In_t r sys -> instance r hh c -> f c -> Forall_t f hh.

  Theorem proof_with_info : forall c, proof sys (projT1 c) -> proof sys' c.
  Proof.
    intros [c ic ] H; simpl in H; generalize c H ic; clear H c ic.
    induction 1 as [ r hh c Hr I Hhh ]; intros Hc.
    specialize (Hf _ _ _ Hr I Hc).
    rewrite <- (Forall_t_lsigT_prop Hf) in I.
    cutrewrite (c = projT1 (existT _ c Hc)) in I; auto.
    apply instance_with_info_prop1 in I.
    apply in_proof with (2 := I).
    unfold sys'; apply In_t_map; auto.
    apply In_t_imp_Forall_t.
    intros [ x Hx ] H.
    apply In_t_map with (f := projT1 (P:=f)) in H.
    rewrite Forall_t_lsigT_prop in H. simpl in H.
    apply (Forall_t_imp_In_t Hhh H Hx).
  Qed.

End stm_with_info.

Implicit Arguments MkRule_info [ stm ].
Implicit Arguments instance_with_info_prop1 [ stm f r hh c ].
Implicit Arguments instance_with_info_prop2 [ stm f r hh c ].
Implicit Arguments proof_forget_info [ stm f sys ].
Implicit Arguments proof_with_info [ stm f sys ].

Section restricted_stm.

  Variable stm : Type.

  Variable P : stm -> Prop.

  Let new_rule r := MkRule (rpred r) (fun p => (lsig_of_lsigT (P:=P) (rhyps r p), sig_of_sigT (rconc r p))).
  
  Definition MkRule_restr r := new_rule (MkRule_info P r).

  Variable sys : list (rule stm).

  Let sys1 := map (MkRule_info P) sys.
  Let sys' := map MkRule_restr sys.

  Let sys'_prop : sys' = map new_rule sys1.
  Proof.
    unfold sys', sys1.
    rewrite map_map.
    apply map_ext.
    intros; auto.
  Qed.

  Let proof_eq_1' : forall c, proof sys1 c -> proof sys' (sig_of_sigT c).
  Proof.
    rewrite sys'_prop.
    apply proof_simul_bij_2.
    intros r hh c I.
    red_instance I p H1 H2 H3.
    unfold new_rule.
    exists p; split; simpl; auto.
    f_equal; symmetry.
    rewrite H3; auto.
    rewrite H2; auto.
  Qed.

  Let proof_eq_'1 : forall c, proof sys' c -> proof sys1 (sigT_of_sig c).
  Proof.
    rewrite sys'_prop.
    apply proof_simul_bij_1.
    intros r hh c I.
    red_instance I p H1 H2 H3.
    prove_instance p; auto.
    rewrite H3, <- map_id.
    unfold lsig_of_lsigT.
    rewrite map_map.
    apply map_ext.
    apply sig_sigT.
    rewrite H2.
    apply sig_sigT.
  Qed.

  Theorem proof_forget_restr : forall c, proof sys' c -> proof sys (proj1_sig c).
  Proof.
    intros c Hc.
    apply proof_eq_'1 in Hc.
    unfold sys1 in Hc.
    apply proof_forget_info in Hc.
    cutrewrite (proj1_sig c = projT1 (sigT_of_sig c)); auto.
    destruct c; auto.
  Qed.

  Hypothesis HP : forall r hh c, In_t r sys -> instance r hh c -> P c -> Forall P hh.

  Theorem proof_with_restr : forall c, proof sys (proj1_sig c) -> proof sys' c.
  Proof.
    intros [ c Hc ] H.
    rewrite <- sigT_sig.
    apply proof_eq_1'; simpl.
    unfold sys1.
    apply (proof_with_info (f:=P)); auto.
    clear c H Hc.
    intros r hh c Hr I Hc.
    apply Forall_Forall_t.
    apply (HP _ _ c Hr); auto.
  Qed.

End restricted_stm.

Implicit Arguments MkRule_restr [ stm ].
Implicit Arguments proof_forget_restr [ stm P sys ].
Implicit Arguments proof_with_restr [ stm P sys ].

    
